#!/bin/sh

B=/hopeless/roots

if [ ! -d $B/$1 ]
then
    echo need base directory
    exit 1
fi

if [ ! -d $B/$2 ]
then
    echo need target directory
    exit 2
fi

mount -t unionfs -o dirs=$B/$2=rw:$B/$1=ro none $B/mnt


