
tar cfz bup-`uname -n`-`date +%Y%m%d%H%M%S`.tar.gz /tftpboot/{pxelinux.0,pxelinux.cfg,bzImage,initrd} /etc/{sysconfig/network,hosts,exports,dhcpd.conf} /hopeless/clients/{template-*,instantiate.sh} /hopeless/{initrd,src} /hopeless/roots/minimal-centos42/imidz.making.of /hopeless/roots/hopeless-minimal-centos42/ /hopeless/roots/mount_union.sh bup.sh doinitrd.sh /etc/rc.d/rc3.d
