#!/bb/bin/sh

export PATH=/bb/bin:/bb/sbin:/bb/usr/bin:/bb/usr/sbin

. /bb/config

echo got $1

if [ "$1" = "bound" ] 
    then
    ifconfig $IFACE $ip netmask $MASK broadcast $BCAST
    hostname $hostname
    echo $ip > /bb/MYIP
    echo $hostname > /bb/HOSTNAME
fi

