#!/bb/bin/sh

export PATH=/bb/bin:/bb/sbin:/bb/usr/bin:/bb/usr/sbin:$PATH

/sbin/insmod /lib/tg3.ko
/sbin/insmod /lib/sunrpc.ko
/sbin/insmod /lib/lockd.ko
/sbin/insmod /lib/nfs.ko
/sbin/insmod /lib/unionfs.ko

#sleep 3
